@@ -1,8 +1,12 @@
 // SPDX-License-Identifier: GPL-2.0
 #include "bcachefs.h"
 
+#include "alloc/accounting.h"
+
 #include "btree/update.h"
 
+#include "fs/namei.h"
+
 #include "init/error.h"
 #include "init/passes.h"
 
@@ -332,15 +336,36 @@ static int snapshot_get_print(struct printbuf *out, struct btree_trans *trans, u
 	if (ret) {
 		prt_str(out, bch2_err_str(ret));
 	} else {
-		if (BCH_SNAPSHOT_SUBVOL(&s))
-			prt_str(out, "subvol ");
-		if (BCH_SNAPSHOT_WILL_DELETE(&s))
-			prt_str(out, "will_delete ");
 		if (BCH_SNAPSHOT_DELETED(&s))
 			prt_str(out, "deleted ");
 		if (BCH_SNAPSHOT_NO_KEYS(&s))
 			prt_str(out, "no_keys ");
-		prt_printf(out, "subvol %u", le32_to_cpu(s.subvol));
+		if (BCH_SNAPSHOT_WILL_DELETE(&s))
+			prt_str(out, "will_delete ");
+		if (BCH_SNAPSHOT_SUBVOL(&s))
+			prt_printf(out, "subvol %u", le32_to_cpu(s.subvol));
+
+		prt_tab(out);
+
+		if (s.subvol) {
+			struct bch_subvolume subvol;
+			ret = lockrestart_do(trans,
+				bch2_subvolume_get(trans, le32_to_cpu(s.subvol), false, &subvol));
+			if (ret)
+				prt_str(out, bch2_err_str(ret));
+			else
+				lockrestart_do(trans,
+					bch2_inum_to_path(trans, (subvol_inum) { le32_to_cpu(s.subvol), le64_to_cpu(subvol.inode) }, out));
+		}
+
+		prt_tab(out);
+
+		u64 v[1] = { 0 };
+		lockrestart_do(trans,
+			       bch2_fs_accounting_read_key2(trans, v, snapshot, id));
+
+		prt_human_readable_u64(out, v[0] << 9);
+		prt_tab_rjust(out);
 	}
 
 	prt_newline(out);
@@ -354,14 +379,18 @@ static unsigned snapshot_tree_max_depth(struct bch_fs *c, u32 start)
 	guard(rcu)();
 	struct snapshot_table *t = rcu_dereference(c->snapshots.table);
 
-	__for_each_snapshot_child(t, start, &depth, id)
+	__for_each_snapshot_child(c, t, start, &depth, id)
 		max_depth = max(depth, max_depth);
 	return max_depth;
 }
 
 int bch2_snapshot_tree_keys_to_text(struct printbuf *out, struct btree_trans *trans, u32 start)
 {
+	printbuf_tabstops_reset(out);
 	printbuf_tabstop_push(out, out->indent + 12 + 2 * snapshot_tree_max_depth(trans->c, start));
+	printbuf_tabstop_push(out, 20);
+	printbuf_tabstop_push(out, 40);
+	printbuf_tabstop_push(out, 12);
 
 	unsigned depth = 0, prev_depth = 0;
 	for_each_snapshot_child(trans->c, start, &depth, id) {
@@ -396,7 +425,7 @@ static u32 bch2_snapshot_right_child(struct snapshot_table *t, u32 id)
 	return bch2_snapshot_child(t, id, 1);
 }
 
-u32 __bch2_snapshot_tree_next(struct snapshot_table *t, u32 id, unsigned *depth)
+u32 __bch2_snapshot_tree_next(struct bch_fs *c, struct snapshot_table *t, u32 id, unsigned *depth)
 {
 	int _depth;
 	if (!depth)
@@ -409,7 +438,7 @@ u32 __bch2_snapshot_tree_next(struct snapshot_table *t, u32 id, unsigned *depth)
 	}
 
 	u32 parent;
-	while ((parent = __bch2_snapshot_parent(t, id))) {
+	while ((parent = __bch2_snapshot_parent(c, t, id))) {
 		(*depth)--;
 		n = bch2_snapshot_right_child(t, parent);
 		if (n && n != id) {
@@ -425,7 +454,7 @@ u32 __bch2_snapshot_tree_next(struct snapshot_table *t, u32 id, unsigned *depth)
 u32 bch2_snapshot_tree_next(struct bch_fs *c, u32 id, unsigned *depth)
 {
 	guard(rcu)();
-	return __bch2_snapshot_tree_next(rcu_dereference(c->snapshots.table), id, depth);
+	return __bch2_snapshot_tree_next(c, rcu_dereference(c->snapshots.table), id, depth);
 }
 
 int bch2_snapshot_lookup(struct btree_trans *trans, u32 id,
@@ -644,3 +673,33 @@ void bch2_fs_snapshots_init_early(struct bch_fs *c)
 
 	mutex_init(&c->snapshots.unlinked_lock);
 }
+
+static int bch2_snapshot_tree_to_text_full(struct printbuf *out, struct btree_trans *trans,
+				    struct bkey_s_c_snapshot_tree st)
+{
+	prt_printf(out, "snapshot tree %llu:\n", st.k->p.offset);
+	guard(printbuf_indent)(out);
+
+	/*
+	 * we need a better way of handling this sort of thing: we don't need
+	 * the outer transaction restart handling loop in for_each_btree_key()
+	 */
+	u32 restart_count = trans->restart_count;
+
+	int ret = bch2_snapshot_tree_keys_to_text(out, trans, le32_to_cpu(st.v->root_snapshot));
+
+	BUG_ON(bch2_err_matches(ret, BCH_ERR_transaction_restart));
+	trans->restart_count = restart_count;
+	return ret;
+}
+
+void bch2_snapshot_trees_to_text(struct printbuf *out, struct bch_fs *c)
+{
+	CLASS(btree_trans, trans)(c);
+	for_each_btree_key(trans, iter,
+				  BTREE_ID_snapshot_trees, POS_MIN, BTREE_ITER_prefetch, k, ({
+		if (k.k->type != KEY_TYPE_snapshot_tree)
+			continue;
+		bch2_snapshot_tree_to_text_full(out, trans, bkey_s_c_to_snapshot_tree(k));
+	}));
+}
