@@ -5,6 +5,8 @@
 #include "btree/key_cache.h"
 #include "btree/update.h"
 
+#include "fs/namei.h"
+
 #include "vfs/fs.h"
 
 #include "init/error.h"
@@ -100,7 +102,8 @@ static int check_subvol(struct btree_trans *trans,
 				subvol_children_iter.pos.inode, subvol_children_iter.pos.offset,
 				(printbuf_reset(&buf),
 				 bch2_bkey_val_to_text(&buf, c, k), buf.buf))) {
-			try(bch2_btree_bit_mod(trans, BTREE_ID_subvolume_children, subvol_children_iter.pos, true));
+			try(bch2_btree_bit_mod(trans, BTREE_ID_subvolume_children,
+					       subvol_children_iter.pos, true));
 		}
 	}
 
@@ -111,9 +114,17 @@ static int check_subvol(struct btree_trans *trans,
 	if (!ret) {
 		if (fsck_err_on(inode.bi_subvol != k.k->p.offset,
 				trans, subvol_root_wrong_bi_subvol,
-				"subvol root %llu:%u has wrong bi_subvol field: got %u, should be %llu",
+				"subvol root %llu:%u has wrong bi_subvol field: got %u, should be %llu\n%s",
 				inode.bi_inum, inode.bi_snapshot,
-				inode.bi_subvol, k.k->p.offset)) {
+				inode.bi_subvol, k.k->p.offset,
+				(printbuf_reset(&buf),
+				 bch2_bkey_val_to_text(&buf, c, k),
+				 prt_newline(&buf),
+				 prt_printf(&buf, "snapshot %u: ", snapid),
+				 bch2_snapshot_to_text(&buf, &snapshot),
+				 prt_newline(&buf),
+				 bch2_inode_unpacked_to_text(&buf, &inode),
+				 buf.buf))) {
 			inode.bi_subvol = k.k->p.offset;
 			inode.bi_snapshot = le32_to_cpu(subvol.snapshot);
 			try(__bch2_fsck_write_inode(trans, &inode));
@@ -659,3 +670,37 @@ void bch2_fs_subvolumes_init_early(struct bch_fs *c)
 	INIT_WORK(&c->snapshots.wait_for_pagecache_and_delete_work,
 		  bch2_subvolume_wait_for_pagecache_and_delete);
 }
+
+static int bch2_subvolume_to_text_full(struct printbuf *out, struct btree_trans *trans, struct bkey_s_c_subvolume s)
+{
+	unsigned pos = out->pos;
+
+	prt_printf(out, "%3llu ", s.k->p.offset);
+
+	int ret = bch2_inum_to_path(trans, (subvol_inum) { s.k->p.offset, le64_to_cpu(s.v->inode) }, out);
+	if (ret) {
+		out->pos = pos;
+		return ret;
+	}
+
+	u32 snapshot = le32_to_cpu(s.v->snapshot);
+
+	prt_printf(out, ":\t%u\r%u\r\n",
+		   bch2_snapshot_tree(trans->c, snapshot),
+		   snapshot);
+	return 0;
+}
+
+void bch2_subvolumes_list_to_text(struct printbuf *out, struct bch_fs *c)
+{
+	printbuf_tabstop_push(out, 40);
+	printbuf_tabstop_push(out, 12);
+	printbuf_tabstop_push(out, 12);
+
+	CLASS(btree_trans, trans)(c);
+	for_each_btree_key(trans, iter, BTREE_ID_subvolumes, POS_MIN, 0, k, ({
+		if (k.k->type != KEY_TYPE_subvolume)
+			continue;
+		bch2_subvolume_to_text_full(out, trans, bkey_s_c_to_subvolume(k));
+	}));
+}
